﻿Imports System.Collections
Imports System.Runtime.CompilerServices

Namespace System.Collections.Generic

   Public Class Heap
      Public Shared Function FromSelector(Of T)(ByVal selector As Func(Of T, IComparable), ByVal minHeap As Boolean) As Heap(Of T)
         Return New Heap(Of T)(selector, minHeap)
      End Function
      Public Shared Function FromSelector(Of T)(ByVal selector As Func(Of T, IComparable), ByVal minHeap As Boolean, ByVal limit As Integer) As LimitedHeap(Of T)
         Return New LimitedHeap(Of T)(selector, minHeap, limit)
      End Function
   End Class

   <Diagnostics.DebuggerDisplay("Count={Count}, MinHeap={MinHeap}")> _
   Public Class Heap(Of T) : Implements ICloneable

      'ein Heap ist ein sortierter, balancierter binärer Baum, im Array (hier: List(of T)) abgebildet.
      'Die beiden Children eines Parents liegen immer genau hinter dem doppelten Index der Parents.
      'Heap-Bedingung (bei MinHeap): beide Children eines Parents müssen größer sein.

      ''' <summary>return 1, 0, -1, depending on key is smaller, equal or bigger than item</summary>
      Public ReadOnly KeyIsSmaller As Func(Of IComparable, T, Integer)
      ''' <summary>return 1, 0, -1, depending on item is smaller, equal or bigger than key</summary>
      Public ReadOnly KeyIsBigger As Func(Of IComparable, T, Integer)
      Public ReadOnly Selector As Func(Of T, IComparable)
      Public ReadOnly MinHeap As Boolean = True
      Protected _InnerList As New List(Of T)()

      Public Sub New(ByVal Selector As Func(Of T, IComparable), ByVal minHeap As Boolean)
         Me.Selector = Selector
         Me.MinHeap = minHeap
         Dim KeyCompareToItem = Function(key As IComparable, item As T) key.CompareTo(Selector(item))
         Dim ItemCompareToKey = Function(key As IComparable, item As T) Selector(item).CompareTo(key)
         KeyIsSmaller = If(minHeap, KeyCompareToItem, ItemCompareToKey)
         KeyIsBigger = If(minHeap, ItemCompareToKey, KeyCompareToItem)
      End Sub
      ''' <summary>copy-constructor</summary>
      Public Sub New(ByVal hp As Heap(Of T))
         MyClass.New(hp.Selector, hp.MinHeap)
         _InnerList.AddRange(hp._InnerList)
      End Sub

      Public Sub Clear()
         _InnerList.Clear()
      End Sub

      ''' <summary>empty the heap and return items in reversed order</summary>
      Public Function FlushToList() As List(Of T)
         Dim heapLength = _InnerList.Count
         For i = heapLength - 1 To 1 Step -1
            Dim itm = _InnerList(i)
            _InnerList(i) = _InnerList(0)
            DownHeap(itm, 0, i)
         Next
         FlushToList = _InnerList
         _InnerList = New List(Of T)
      End Function

      ''' <summary>return sorted items </summary>
      Public Function ToList() As List(Of T)
         Dim hp As New Heap(Of T)(Me)
         Dim lst = hp.FlushToList
         lst.Reverse()
         Return lst
      End Function

      ''' <summary>Insert Item</summary>
      Public Overridable Sub Push(ByVal item As T)
         _InnerList.Add(Nothing)      ' leeren Platz am Ende schaffen
         _InnerList(UpHeap(item, _InnerList.Count - 1)) = item
      End Sub

      Protected Function UpHeap(ByVal item As T, ByVal iChild As Integer) As Integer
         Dim key = Selector(item)
         'shift the parent on current child-position, until parent should be items parent, or position 0 is reached
         While iChild <> 0
            Dim iParent = (iChild - 1) >> 1
            If KeyIsSmaller(key, _InnerList(iParent)) >= 0 Then Exit While 'Parent ist kleiner
            _InnerList(iChild) = _InnerList(iParent)         'bisherigen Parent auf iChild schreiben
            iChild = iParent
         End While
         Return iChild
      End Function

      Protected Sub DownHeap(ByVal item As T, ByVal iParent As Integer, ByVal heapLength As Integer)
         Dim key = Selector(item)
         'shift the higher-priority-child on current parent-position, until item should be childs parent, or heapLength is reached
         Dim iChild = (iParent << 1) + 1
         While iChild < heapLength
            ' Heap - binärer Baum im Array: die beiden Children eines Knotens
            ' liegen genau hinter der Verdopplung des Parent-Index'
            Dim child = _InnerList(iChild)
            Dim iChild2 = iChild + 1
            If iChild2 < heapLength Then
               Dim child2 = _InnerList(iChild2)
               If KeyIsBigger(Selector(child2), child) > 0 Then
                  iChild = iChild2
                  child = child2
               End If
            End If
            If KeyIsBigger(key, child) >= 0 Then Exit While
            _InnerList(iParent) = child
            iParent = iChild
            iChild = (iParent << 1) + 1
         End While
         _InnerList(iParent) = item
      End Sub

      ''' <summary>remove Top-Item and return it</summary>
      Public Overridable Function Pop() As T
         Pop = _InnerList(0)
         Dim heapLength = _InnerList.Count - 1
         Dim item = _InnerList(heapLength)
         DownHeap(item, 0, heapLength)
         _InnerList.RemoveAt(heapLength)
      End Function

      ''' <summary>return Top-Item without remove it</summary>
      Public Function Peek() As T
         Return _InnerList(0)
      End Function

      Public ReadOnly Property Count() As Integer
         Get
            Return _InnerList.Count
         End Get
      End Property

      ''' <summary>if items Priority &lt; TopItem - incorporate item and "ByRef-return" previous TopItem</summary>
      ''' <returns>whether incorporation was executed</returns>
      Public Function TryExChange(ByRef item As T) As Boolean
         If KeyIsBigger(Selector(item), _InnerList(0)) >= 0 Then Return False
         item = ExChangeTopItem(item)
         Return True
      End Function

      ''' <summary>incorporate item and return previous TopItem</summary>
      Public Function ExChangeTopItem(ByVal item As T) As T
         ExChangeTopItem = _InnerList(0)
         DownHeap(item, 0, _InnerList.Count)
      End Function

      ''' <summary>Rebalance heap if necessary after TopNodes weighting changed </summary>
      Public Sub ReHeap()
         Dim item = _InnerList(0)
         DownHeap(item, 0, _InnerList.Count)
      End Sub

      ''' <summary>
      ''' Rebalances Heap after changing Priority of an Item. Note: ChangeKey() always fails with items of immutable or Structure-Type 
      ''' </summary>
      ''' <exception cref="Exception(Of Heap(of T))">Heap doesnt contain Item</exception>
      Public Sub ChangeKey(ByVal item As T)
         Dim i = _InnerList.IndexOf(item) 'imperformant, but array-index-tracking is still more imperformant
         If i < 0 Then Throw Me.Exception("ChangeKey(", GetType(T).Name, _
               "): item could not be found in Heap. Note: ChangeKey() always fails with items of immutable or Structure-Type")
         i = UpHeap(item, i)
         DownHeap(item, i, _InnerList.Count)
      End Sub

      Public Overridable Function Clone() As Object Implements ICloneable.Clone
         Dim res = New Heap(Of T)(Selector, MinHeap)
         res._InnerList.AddRange(_InnerList)
         Return res
      End Function
   End Class

   ''' <summary>note: limited MinHeap selects biggest values</summary>
   Public Class LimitedHeap(Of T) : Inherits Heap(Of T)
      Public ReadOnly Limit As Integer
      Public Sub New(ByVal Selector As Func(Of T, IComparable), ByVal minHeap As Boolean, ByVal limit As Integer)
         MyBase.New(Selector, minHeap)
         Me.Limit = limit
      End Sub
      Public Overrides Sub Push(ByVal item As T)
         If Count < Limit Then
            MyBase.Push(item)
         Else
            If KeyIsBigger(Selector(item), _InnerList(0)) >= 0 Then Return 'ablehnen, wenn item kleiner
            DownHeap(item, 0, _InnerList.Count)
         End If
      End Sub
      Public Overrides Function Clone() As Object
         Dim res = New LimitedHeap(Of T)(Selector, MinHeap, Limit)
         res._InnerList.AddRange(_InnerList)
         Return res
      End Function
   End Class
End Namespace